from collections import defaultdict
from typing import List, Optional

import idaapi

from threatray_ida_plugin.libs.api.api import API
from threatray_ida_plugin.libs.api.api_error import FailedAPIRequest
from threatray_ida_plugin.libs.logger import get_log
from threatray_ida_plugin.libs.utils import get_hash_sha256_of_input_file
from threatray_ida_plugin.models.function_retrohunt_result import FunctionRetrohuntCodeRegionWithFunctions
from threatray_ida_plugin.views.function_retrohunt.function_retrohunt_result_view import FunctionRetrohuntResultView
from threatray_ida_plugin.views.function_retrohunt.function_retrohunt_threshold_box import FunctionRetrohuntThresholdBox
from threatray_ida_plugin.views.message_box import MessageBox

log = get_log()

FEATURE_NAME: str = 'Function retrohunt'
INPUT_FUNCTION_LIMIT: int = 100


class FunctionRetrohuntService:
    def __init__(self, api: API):
        self.__api = api

    def _get_threshold(self, number_of_input_functions: int) -> Optional[int]:
        if number_of_input_functions == 1:
            threshold = 1
        else:
            threshold = FunctionRetrohuntThresholdBox.create(number_of_input_functions)
            if threshold is None:  # canceled
                log.debug(f'{FEATURE_NAME} process canceled.')
            elif not threshold:
                msg = f'Invalid threshold. {FEATURE_NAME} process canceled.'
                MessageBox.create('Warning', msg)
                log.warning(msg)
        return threshold

    def workflow(self, selected_functions: List[int]):
        if len(selected_functions) > INPUT_FUNCTION_LIMIT:
            msg = (f'Only {INPUT_FUNCTION_LIMIT} input functions are allowed (currently {len(selected_functions)} '
                   f'functions are selected). {FEATURE_NAME} process canceled.')
            MessageBox.create('Error - Too many input functions', msg)
            log.error(msg)
            return

        hash_sha256 = get_hash_sha256_of_input_file()
        if 'No such file or directory' in hash_sha256:
            MessageBox.create('Warning - File not found', hash_sha256)
            log.warning(hash_sha256)
            return

        if not selected_functions:
            msg = f'No function selected. {FEATURE_NAME} process canceled.'
            MessageBox.create('Warning - No function selected', msg)
            log.warning(msg)
            return

        threshold = self._get_threshold(len(selected_functions))
        if not threshold:  # None or 0 is both invalid
            return

        functions_by_hash = self.__api.get_functions(hash_sha256)
        if isinstance(functions_by_hash, FailedAPIRequest):
            msg = f'{functions_by_hash.message}. {FEATURE_NAME} process canceled'
            MessageBox.create('Warning - Failed Api Request', msg)
            log.warning(msg)
            return
        input_functions = list({func['uid'] for func in functions_by_hash.get('functions', [])
                                if func['address'] in selected_functions})
        if not input_functions:
            msg = f'No input functions found. {FEATURE_NAME} process canceled.'
            MessageBox.create('Warning', msg)
            log.warning(msg)
            return

        log.debug(f'{FEATURE_NAME} with {len(input_functions)} unique functions and a threshold of {threshold}.')
        log.info('Doing threatray function retrohunt...')
        idaapi.show_wait_box('Doing threatray function retrohunt...')

        result = self.__api.retrohunt_by_function(input_functions, threshold)
        if isinstance(result, FailedAPIRequest):
            msg = f'{result.message}. {FEATURE_NAME} process canceled'
            MessageBox.create('Warning - Failed Api Request', msg)
            log.warning(msg)
            idaapi.hide_wait_box()
            return
        idaapi.hide_wait_box()
        log.info(f'Found {len(result)} matching binary files.')

        uid_to_addresses = defaultdict(list)
        for function in functions_by_hash['functions']:
            uid_to_addresses[function['uid']].append(hex(function['address']))
        final_result = [FunctionRetrohuntCodeRegionWithFunctions.from_function_retrohunt_code_region_with_uids(
            function_retrohunt_with_uids=r,
            matching_input_functions=[address for uid in r.matching_uids for address in uid_to_addresses[uid]]
        ) for r in result]
        FunctionRetrohuntResultView.create(final_result)
