from collections import Counter
from typing import List, Dict, Optional

import idaapi
import idautils
import idc

from threatray_ida_plugin.libs.api.api import API
from threatray_ida_plugin.libs.api.api_error import FailedAPIRequest
from threatray_ida_plugin.libs.logger import get_log
from threatray_ida_plugin.libs.utils import get_hash_sha256_of_input_file
from threatray_ida_plugin.models.signature_inspection_result import SignatureInspectionResult, Threat
from threatray_ida_plugin.views.message_box import MessageBox
from threatray_ida_plugin.views.signature_inspection.signature_inspection_color_selection_box import \
    SignatureInspectionColorSelectionBox
from threatray_ida_plugin.views.signature_inspection.signature_inspection_result_view import \
    SignatureInspectionResultView
from threatray_ida_plugin.views.signature_inspection.signature_inspection_setting_box import \
    SignatureInspectionSettingBox

log = get_log()

FEATURE_NAME: str = 'Signature inspection'
COMMENT_PREFIX: str = 'Threatray Threats: '

COLOR_PALETTE: List[int] = [10079487, 10066431, 10092543, 10092492, 10092441, 13434777, 16777113, 16764057, 16751001,
                            16751052, 16751103, 13408767]


def set_function_color(address: int, color: int):
    idc.set_color(address, idc.CIC_FUNC, color)


def set_xref_color(address: int, color: int):
    idc.set_color(address, idc.CIC_ITEM, color)


class SignatureInspectionService:
    def __init__(self, api: API):
        self.__api = api

    def __annotate_comments(self, result: List[SignatureInspectionResult]):
        # Note: If there are non-repeatable comments, they can be overwritten or hidden after threatray comments are set
        for function in result:
            threats = [threat.label for threat in function.threats]
            if not threats:
                # don't set a comment when there are no threats
                continue
            comment = f'{COMMENT_PREFIX}{threats}'
            func = idaapi.get_func(int(function.address))
            if not func:
                log.debug(f'Could not find function for result: {function}')
                continue
            current_repeatable_comment = idaapi.get_func_cmt(func, True)
            if current_repeatable_comment:
                for current_comment_line in current_repeatable_comment.splitlines():
                    if current_comment_line.startswith(COMMENT_PREFIX):
                        continue
                    comment = f'{comment}\n{current_comment_line}'
            idaapi.set_func_cmt(func, comment, True)

    def __apply_color_to_functions_and_their_xrefs(self, functions_to_color: Dict[int, int]):
        for function_address, color in functions_to_color.items():
            set_function_color(function_address, color)
            for xref in idautils.XrefsTo(function_address):
                set_xref_color(xref.frm, color)

    def __color_functions_and_their_xrefs(self, result: List[SignatureInspectionResult]):
        function_address_to_color: Dict[int, int] = {}
        threat_to_color: Dict[str, int] = {}
        color_index: int = 0

        threat_prevalence = get_threat_prevalence(result)
        for function in result:
            if not idaapi.get_func(function.address):
                log.debug(f'Could not find function: {function}')
                continue
            most_prevalent_threat = get_most_prevalent_threat(function, threat_prevalence)
            color = idc.DEFCOLOR  # white
            if most_prevalent_threat:
                if most_prevalent_threat.label not in threat_to_color:
                    threat_to_color[most_prevalent_threat.label] = COLOR_PALETTE[color_index]
                    # we reuse color if there are a lot of different threats
                    color_index = (color_index + 1) % len(COLOR_PALETTE)
                color = threat_to_color[most_prevalent_threat.label]
            function_address_to_color[function.address] = color
        self.__apply_color_to_functions_and_their_xrefs(function_address_to_color)

    def __color_functions_and_their_xrefs_from_manual_selection(self, result: List[SignatureInspectionResult],
                                                                threat_label_to_color: Dict[str, int]):
        function_address_to_color: Dict[int, int] = {}
        threat_prevalence = get_threat_prevalence(result)

        # The color of the first threat with a selected color (!= DEFCOLOR) is used for each function.
        # If several threats of a function have a selected color, the color of the most prevalent threat is selected.
        for function in result:
            if not idaapi.get_func(function.address):
                log.debug(f'Could not find function: {function}')
                continue
            most_prevalent_threat = get_most_prevalent_threat(function, threat_prevalence)
            color = idc.DEFCOLOR  # white
            for threat in function.threats:
                if threat_label_to_color[threat.label] == idc.DEFCOLOR:
                    continue
                if color != idc.DEFCOLOR and threat.label != most_prevalent_threat.label:
                    continue
                color = threat_label_to_color[threat.label]
            function_address_to_color[function.address] = color
        self.__apply_color_to_functions_and_their_xrefs(function_address_to_color)

    def workflow(self):
        hash_sha256 = get_hash_sha256_of_input_file()
        if 'No such file or directory' in hash_sha256:
            MessageBox.create('Warning - File not found', hash_sha256)
            log.warning(hash_sha256)
            return

        settings = SignatureInspectionSettingBox.create()
        if settings is None:  # canceled
            log.debug(f'{FEATURE_NAME} process canceled.')
            return

        msg = 'Getting Signature Inspection results...'
        log.info(msg)
        idaapi.show_wait_box(msg)
        result = self.__api.get_signature_inspection(hash_sha256)
        idaapi.hide_wait_box()

        if isinstance(result, FailedAPIRequest):
            msg = f'{result.message}. {FEATURE_NAME} canceled'
            MessageBox.create('Warning - Failed Api Request', msg)
            log.warning(msg)
            return

        self.__annotate_comments(result)

        if settings.color_functions_with_hit:
            if settings.manually_select_signature_hits_to_be_colored:
                threat_labels = {threat.label for r in result for threat in r.threats}
                threat_label_to_color = SignatureInspectionColorSelectionBox.create(threat_labels)
                if threat_label_to_color is None:  # canceled
                    log.debug(f'{FEATURE_NAME} process canceled.')
                    return
                self.__color_functions_and_their_xrefs_from_manual_selection(result, threat_label_to_color)
            else:
                self.__color_functions_and_their_xrefs(result)
        SignatureInspectionResultView.create(result, settings=settings)


def get_threat_prevalence(result: List[SignatureInspectionResult]) -> Counter[str, int]:
    threat_prevalence = Counter()
    for r in result:
        for threat in r.threats:
            threat_prevalence[threat.label] += 1
    return threat_prevalence


def get_most_prevalent_threat(result: SignatureInspectionResult,
                              threat_prevalence: Counter[str, int]) -> Optional[Threat]:
    # return the most prevalent threat; if two threats are equally prevalent, select the first one
    if not result.threats:
        return None
    selected_threat = result.threats[0]
    for threat in result.threats[1:]:
        if threat_prevalence.get(threat.label, 1) > threat_prevalence.get(selected_threat.label, 1):
            selected_threat = threat
    return selected_threat
