from typing import List

from threatray_ida_plugin.libs.api.api import API
from threatray_ida_plugin.libs.api.api_error import FailedAPIRequest
from threatray_ida_plugin.libs.logger import get_log
from threatray_ida_plugin.libs.utils import get_hash_sha256_of_input_file
from threatray_ida_plugin.views.message_box import MessageBox
from threatray_ida_plugin.views.signature_management.signature_action_box import SignatureActionBox, SignatureAction
from threatray_ida_plugin.views.signature_management.signature_creation_threshold_box import \
    SignatureCreationThresholdBox
from threatray_ida_plugin.views.signature_management.signature_label_box import SignatureLabelBox

log = get_log()
FEATURE_NAME: str = 'Update signature'


class SignatureManagementService:
    def __init__(self, api: API):
        self.__api = api

    def workflow(self, selected_functions: List[int]):
        hash_sha256 = get_hash_sha256_of_input_file()
        if 'No such file or directory' in hash_sha256:
            MessageBox.create('Warning - File not found', hash_sha256)
            log.warning(hash_sha256)
            return

        if not selected_functions:
            msg = f'No function selected. {FEATURE_NAME} process canceled.'
            MessageBox.create('Warning - No function selected', msg)
            log.warning(msg)
            return

        signatures = self.__api.get_signatures()
        if isinstance(signatures, FailedAPIRequest):
            msg = f'{signatures.message}. {FEATURE_NAME} process canceled'
            MessageBox.create('Warning - Failed Api Request', msg)
            log.warning(msg)
            return

        signature_label = SignatureLabelBox.create(labels=sorted(signature.label for signature in signatures))
        if signature_label is None:  # canceled
            log.debug(f'{FEATURE_NAME} process canceled.')
            return
        if not signature_label:
            msg = f'Signature label missing. {FEATURE_NAME} process canceled.'
            MessageBox.create('Warning - Signature label missing', msg)
            log.warning(msg)
            return

        signature = self.__api.get_signature(signature_label)
        if isinstance(signature, FailedAPIRequest):
            msg = f'{signature.message}. Update signature process canceled'
            MessageBox.create('Warning - Failed Api Request', msg)
            log.warning(msg)
            return

        if not signature:
            signature_threshold_response = SignatureCreationThresholdBox.create(len(selected_functions))
            if signature_threshold_response is None:
                log.warning(f'{FEATURE_NAME} process canceled')
                return

            signature = self.__api.create_signature(signature_label,
                                                    signature_threshold_response.threshold_suspicious,
                                                    signature_threshold_response.threshold_malicious)
            if isinstance(signature, FailedAPIRequest):
                msg = f'Failed to create signature: {signature.message}. {FEATURE_NAME} process canceled'
                MessageBox.create('Warning - Failed Api Request', msg)
                log.warning(msg)
                return

            log.info(f'Created signature {signature_label}')

            action = SignatureAction.ADD.value  # it is a new signature, so only adding functions makes sense
        else:
            log.info(f'Selected existing signature {signature_label}')
            action = SignatureActionBox.create()
            if action is None:  # canceled
                log.debug(f'{FEATURE_NAME} process canceled.')
                return
            try:
                action = SignatureAction(action).value
            except ValueError as _:
                msg = f'No action selected. {FEATURE_NAME} process canceled.'
                MessageBox.create('Warning', msg)
                log.warning(msg)
                return

        functions_by_hash = self.__api.get_functions(hash_sha256)
        if isinstance(functions_by_hash, FailedAPIRequest):
            msg = f'{functions_by_hash.message}. {FEATURE_NAME} process canceled'
            MessageBox.create('Warning - Failed Api Request', msg)
            log.warning(msg)
            return

        analysis_id = functions_by_hash['analysis']['analysis_id']
        pid = functions_by_hash['code_region']['pid']
        base = functions_by_hash['code_region']['base']
        code_region_hash = functions_by_hash['code_region']['hash_sha256']

        functions_to_update = [{
            'uid': func['uid'],
            'address': func['address'],
            'analysis_id': analysis_id,
            'pid': pid,
            'base': base,
            'code_region_hash': code_region_hash
        } for func in functions_by_hash.get('functions', []) if func['address'] in selected_functions]

        if not functions_to_update:
            msg = f'No functions found to update the signature. {FEATURE_NAME} process canceled.'
            MessageBox.create('Warning', msg)
            log.warning(msg)
            return

        log.debug(f"{action.capitalize()}ing the following functions of analysis {analysis_id}, "
                  f" pid {pid}, base {base}, and hash {code_region_hash} to {signature_label}")
        for function in functions_to_update:
            log.debug(f'{function["uid"]}/{function["address"]}')

        result = self.__api.update_signature(signature.id, functions_to_update, action=action)
        if isinstance(result, FailedAPIRequest):
            msg = f'{result.message}. {FEATURE_NAME} process failed.'
            MessageBox.create('Warning - Failed Api Request', msg)
            log.warning(msg)
        else:
            msg = (f'The selected functions were {str(action)}ed to the signature {signature_label}.\n'
                   f'It may take a while until all signature results are updated.')
            MessageBox.create('Info', msg)
            log.info(msg)
