import json
from pathlib import Path
from typing import Optional, Union, Dict, List

import requests

from threatray_ida_plugin.libs.api.api_error import APIError, FailedAPIRequest, APISetupError
from threatray_ida_plugin.libs.logger import get_log
from threatray_ida_plugin.libs.oauth2 import Oauth2
from threatray_ida_plugin.models.function_retrohunt_result import FunctionRetrohuntCodeRegionWithUids
from threatray_ida_plugin.models.signature import Signature
from threatray_ida_plugin.models.signature_inspection_result import SignatureInspectionResult, Threat

log = get_log()
DEFAULT_SECRETS_FILE = 'threatray_api_secrets.json'


def _parse_response(r: requests.Response) -> Dict:
    try:
        data = r.json()
    except:
        log.debug(f'resp.status_code={r.status_code} - resp.content={r.content}')
        raise APIError('Invalid API response.')

    if r.status_code == 404:
        log.debug(f'resp.content={r.content}')
        raise APIError('No entity found with this ID/Hash.')

    if r.status_code == 400:
        log.debug(f'resp.content={r.content}')
        raise APIError('Invalid request.')

    if not r.ok:
        log.debug(f'resp.status_code={r.status_code} - resp.content={r.content}')
        raise APIError(data.get('description', 'Unknown error.'))

    return data


class API:
    def __init__(self):
        secrets_config: Optional[Dict] = get_secrets_config()
        if not secrets_config:
            raise APISetupError(f'Secrets file {DEFAULT_SECRETS_FILE} is invalid.')

        self.__oauth2 = Oauth2(secrets_config['realm'])
        self.__oauth2.set_secret_token(secrets_config['secret_token'])

    def __endpoint(self, endpoint: str) -> str:
        return f'{self.__oauth2.get_api_url()}/v1/{endpoint}'

    def get_functions(self, hash_sha256: str) -> Union[Dict, FailedAPIRequest]:
        params = {'selector': {'hash': hash_sha256}}
        resp: requests.Response = requests.get(
            self.__endpoint('functions'),
            json=params,
            headers=self.__oauth2.get_authorization_header()
        )

        try:
            return _parse_response(resp)
        except APIError as e:
            return FailedAPIRequest(message=str(e))

    def get_signatures(self) -> Union[List[Signature], FailedAPIRequest]:
        resp: requests.Response = requests.get(
            self.__endpoint('signatures'),
            headers=self.__oauth2.get_authorization_header()
        )

        try:
            return [Signature(id=signature['id'],
                              label=signature['label']) for signature in _parse_response(resp).get('signatures')]
        except APIError as e:
            log.debug(e)
            return FailedAPIRequest(message=str(e))

    def get_signature_inspection(self, hash_sha256: str) -> Union[List[SignatureInspectionResult], FailedAPIRequest]:
        params = {'hash': hash_sha256}
        # TODO: switch to signature-inspection, must make sure that rebasing is managed correctly
        # TODO: do we always rebase? what if we only send hash (no base), what happens then?
        resp: requests.Response = requests.post(
            self.__endpoint('signature-inspection'),
            json=params,
            headers=self.__oauth2.get_authorization_header()
        )
        try:
            result = _parse_response(resp)
            return [SignatureInspectionResult(
                address=function['address'],
                size=function['size'],
                threats=[Threat(label=threat['label'],
                                sample=threat['sample'],
                                binary_file=threat['code_region'],
                                address=threat['address'])
                         for threat in function['threats']]
            ) for function in result['functions']]
        except APIError as e:
            return FailedAPIRequest(message=str(e))

    def __retrohunt(self, functions_uids: List, threshold: float, scope: str = 'both') -> Union[Dict, FailedAPIRequest]:
        resp: requests.Response = requests.get(
            self.__endpoint('retrohunt/functions'),
            json={'uids': functions_uids,
                  'scope': scope,
                  'threshold': threshold},
            headers=self.__oauth2.get_authorization_header()
        )
        return _parse_response(resp)

    def retrohunt_by_function(self, functions_uids: List,
                              threshold: int,
                              scope: str = 'both') -> Union[
        List[FunctionRetrohuntCodeRegionWithUids], FailedAPIRequest]:
        results = {}
        total_uids = len(functions_uids)
        percentage_threshold = min(float(threshold / total_uids), 1.0)

        while functions_uids:
            try:
                retrohunt_results = self.__retrohunt(functions_uids[:100], percentage_threshold, scope)
            except APIError as e:
                return FailedAPIRequest(message=str(e))

            analysis_id_to_analysis = {a['analysis_id']: a for a in retrohunt_results['analyses']}
            functions_uids = functions_uids[100:]

            for code_region in retrohunt_results['code_regions']:
                analysis = analysis_id_to_analysis.get(code_region['analysis_id'])
                if not analysis:
                    # Be defensive
                    continue

                cr = FunctionRetrohuntCodeRegionWithUids.from_dict(code_region, analysis['created_at'])
                key = cr.get_key()
                if key in results:
                    results[key] = results[key].add_nr_of_function_matches(cr.nr_of_function_matches)
                else:
                    results[key] = cr

            # register uids
            for function in retrohunt_results['functions']:
                for cr_match in function['matches']:
                    key = f'{cr_match["analysis_id"]}_{cr_match["pid"]}_{cr_match["base"]}_{cr_match["code_region"]}'
                    results[key] = results[key].add_matching_uid(function['uid'])

        return [result for result in sorted(results.values(), key=lambda r: -r.nr_of_function_matches)
                if result.nr_of_function_matches >= threshold]

    def get_signature(self, signature_label: str) -> Union[Optional[Signature], FailedAPIRequest]:
        response = self.get_signatures()
        if isinstance(response, FailedAPIRequest):
            return response

        for signature in response:
            if signature.label.lower() == signature_label.lower():
                return signature
        return None

    def create_signature(self, signature_label: str,
                         threshold_suspicious: int,
                         threshold_malicious: int
                         ) -> Union[Signature, FailedAPIRequest]:
        resp: requests.Response = requests.post(
            self.__endpoint('signatures'),
            json={'label': signature_label,
                  "code_classifier_config": {
                      "threshold_suspicious": threshold_suspicious,
                      "threshold_malicious": threshold_malicious
                  },
                  },
            headers=self.__oauth2.get_authorization_header()
        )

        try:
            data = _parse_response(resp)
            return Signature(id=data['id'], label=data['label'])
        except APIError as e:
            return FailedAPIRequest(message=str(e))

    def update_signature(self, signature_id: int,
                         functions: List[Dict],
                         action: str) -> Union[Dict, FailedAPIRequest]:
        if action not in ['add', 'remove']:
            return FailedAPIRequest(message=f'Invalid action {action}.')

        resp: requests.Response = requests.post(
            self.__endpoint(f'signatures/{signature_id}/functions'),
            json={'functions': [{
                'action': action,
                'address': function['address'],
                'analysis_id': function['analysis_id'],
                'pid': function['pid'],
                'base': function['base'],
                'code_region_hash': function['code_region_hash'],
            } for function in functions]},
            headers=self.__oauth2.get_authorization_header()
        )

        try:
            return _parse_response(resp)
        except APIError as e:
            return FailedAPIRequest(message=str(e))


def get_secrets_config() -> Optional[Dict]:
    file_path: Path = Path(__file__).resolve().parent / Path(DEFAULT_SECRETS_FILE)
    if not file_path.exists():
        # Check home directory
        home_path: Path = Path().home() / '.threatray' / DEFAULT_SECRETS_FILE
        if home_path.exists():
            file_path = home_path

    if not Path(file_path).exists():
        raise APISetupError(f'Did not find {DEFAULT_SECRETS_FILE} in {file_path.parent} or '
                            'in ~/.threatray. Did you create it with: generate_api_secrets.py?')

    if Path(file_path).is_file():
        with open(file_path, 'r') as f:
            secrets = json.load(f)
            if secrets.get('realm') and secrets.get('secret_token'):
                return secrets

    return None
