import logging
from typing import Dict

import requests
from authlib.jose import jwt, errors

logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s %(levelname)s:%(name)s:%(filename)s:L%(lineno)d:%(message)s'
)
log = logging.getLogger('auth.oauth2')


class Credential:
    def __init__(self, username, password):
        self.username = username
        self.password = password


class Oauth2:
    __client_id = 'threatray-ui'
    __token = None
    __certs = None
    __credential = None
    __offline_token = None

    def __init__(self, realm: str, context: str = 'prod', verbose: bool = False):
        self.realm = realm
        self.context = context
        self.verbose = verbose

    def set_credential(self, credential: Credential):
        self.__credential = credential

    def set_secret_token(self, offline_token: str):
        self.__offline_token = offline_token

    def get_access_token(self) -> str:
        valid = False
        if self.__token and self.__is_token_valid(self.__token['access_token']):
            valid = True
        elif self.__token:
            try:
                self.__refresh_access_token(self.__token['refresh_token'])
                valid = True
            except Exception as e:
                valid = False
                if self.verbose:
                    log.debug(e)

        if not valid:
            if self.__credential:
                self.__password_auth()
            elif self.__offline_token:
                self.__refresh_access_token(self.__offline_token)
            else:
                log.error('No credentials or offline token configured')

        return self.__token['access_token']

    def get_secret_token(self, credential: Credential) -> str:
        data = {
            'grant_type': 'password',
            'client_id': self.__client_id,
            'username': credential.username,
            'password': credential.password,
            'scope': 'offline_access'
        }
        req = requests.post(self.__get_token_url(), data=data)
        if req.ok:
            return req.json().get('refresh_token')

        msg = f'Failed to fetch OAUTH2_TOKEN: {req.content}'
        if self.verbose:
            log.error(msg)

        raise Exception(msg)

    def __password_auth(self):
        data = {
            'grant_type': 'password',
            'client_id': self.__client_id,
            'username': self.__credential.username,
            'password': self.__credential.password
        }
        req = requests.post(self.__get_token_url(), data=data)
        if req.ok:
            self.__token = req.json()
        else:
            msg = f'Failed to fetch OAUTH2_TOKEN: {req.content}'
            if self.verbose:
                log.error(msg)
            raise Exception(msg)

    def __refresh_access_token(self, refresh_token: str):
        data = {
            'grant_type': 'refresh_token',
            'client_id': self.__client_id,
            'refresh_token': refresh_token
        }
        req = requests.post(self.__get_token_url(), data=data)
        if req.ok:
            self.__token = req.json()
        else:
            msg = f'Failed to re-fresh OAUTH2_TOKEN: {req.content}'
            if self.verbose:
                log.error(msg)
            raise Exception(msg)

    def __is_token_valid(self, token) -> bool:
        try:
            jwt.decode(token, self.__get_keycloak_certs()).validate()
            return True
        except errors.ExpiredTokenError:
            pass  # OK, will be refreshed
        except Exception as e:
            if self.verbose:
                msg = f'Invalid token: {e}'
                log.error(msg)
        return False

    def __get_keycloak_certs(self) -> Dict:
        if not self.__certs:
            url = f'{self.get_auth_url()}/auth/realms/{self.realm}/protocol/openid-connect/certs'

            req = requests.get(url)
            if not req.ok:
                msg = f'Failed to fetch public key: {req}'
                if self.verbose:
                    log.error(msg)
                raise Exception(msg)
            self.__certs = req.json()
        return self.__certs

    def __get_token_url(self) -> str:
        return f'{self.get_auth_url()}/auth/realms/{self.realm}/protocol/openid-connect/token'

    def get_authorization_header(self) -> Dict:
        return {'Authorization': f'Bearer {self.get_access_token()}'}

    def get_auth_url(self) -> str:
        return f'https://login.{"analysis" if self.context == "prod" else "dev"}.threatray.com'

    def get_api_url(self) -> str:
        return f'https://api-{self.realm}.{"analysis" if self.context == "prod" else "dev"}.threatray.com'
