from typing import Optional

import idaapi

from threatray_ida_plugin.libs.constants import OKAY_RESPONSE
from threatray_ida_plugin.libs.logger import get_log

log = get_log()

TEXT = r"""BUTTON YES* Select
BUTTON NO NONE
BUTTON CANCEL Cancel
Select Function Retrohunt Threshold

{form_change}
<Threshold:{threshold}>
It is the number of input function that must match with a binary file for it to be returned
"""

MINIMAL_THRESHOLD: int = 1


class FunctionRetrohuntThresholdBox(idaapi.Form):
    def __init__(self, number_of_selected_functions: int):
        self.__number_of_selected_functions = number_of_selected_functions
        self.canceled: bool = False
        self.selected_threshold: int = max(int(self.__number_of_selected_functions * 0.2), MINIMAL_THRESHOLD)

        idaapi.Form.__init__(
            self,
            TEXT,
            {
                'form_change': idaapi.Form.FormChangeCb(self.on_form_change),
                'threshold': idaapi.Form.NumericInput(value=self.selected_threshold, tp=idaapi.Form.FT_DEC)
            }
        )

    def on_form_change(self, fid):
        try:
            threshold = int(self.GetControlValue(self.threshold))
            if threshold > self.__number_of_selected_functions:
                log.warning(f'Threshold should not be greater than the number of input functions '
                            f'(={self.__number_of_selected_functions}).')
                self.selected_threshold = self.__number_of_selected_functions
            elif threshold < 1:
                log.warning('Threshold must be a positive integer.')
                self.selected_threshold = MINIMAL_THRESHOLD
            else:
                self.selected_threshold = threshold
        except (ValueError, TypeError) as e:
            log.error(f'Invalid threshold provided: {e}')

        # we update the threshold in the UI in case a float number was provided or an error has happened
        self.SetControlValue(self.threshold, self.selected_threshold)
        if fid == idaapi.CB_CANCEL:
            self.canceled = True
        return OKAY_RESPONSE

    @staticmethod
    def create(number_of_selected_functions: int) -> Optional[int]:
        box = FunctionRetrohuntThresholdBox(number_of_selected_functions)
        box.Compile()
        box.Execute()
        if box.canceled:
            return None
        return box.selected_threshold
