from typing import Optional, Dict, Set

import idaapi
import idc

from threatray_ida_plugin.libs.constants import OKAY_RESPONSE
from threatray_ida_plugin.libs.logger import get_log

log = get_log()

TEXT = r"""BUTTON YES* Select
BUTTON NO NONE
BUTTON CANCEL Cancel
Signature Inspection Color Settings

Please select the desired color for each signature. 
DEFAULT means that the hits of the signature are not colored:
{form_change}
"""
NUMBER_OF_COLUMNS: int = 2
WHITE_COLOR: hex = 0xFFFFFF


class SignatureInspectionColorSelectionBox(idaapi.Form):
    def __init__(self, threat_labels: Set[str]):
        self.canceled: bool = False
        self.__threat_label_to_color_input: Dict[str, idaapi.Form.ColorInput] = {
            threat_label: idaapi.Form.ColorInput(idc.DEFCOLOR) for threat_label in threat_labels}
        self.threat_label_to_color: Dict[str, int] = {threat_label: idc.DEFCOLOR for threat_label in threat_labels}
        self.__offset = 0

        text = TEXT
        largest_threat_label = max(len(label) for label in threat_labels)
        sorted_color_picker = sorted(self.__threat_label_to_color_input.items(), key=lambda item: item[0])
        for i, (threat_label, color_input) in enumerate(sorted_color_picker, start=1):
            space = " " * (largest_threat_label - len(threat_label))
            text += f"<Signature {threat_label}{space}:" + "{" + f"{threat_label}" + "}>"
            text += "\n" if i % NUMBER_OF_COLUMNS == 0 else "\t"
        controls = {
            'form_change': idaapi.Form.FormChangeCb(self.on_form_change),
        }
        controls.update(self.__threat_label_to_color_input)
        idaapi.Form.__init__(self, text, controls)

    def on_form_change(self, fid):
        for threat_label, color_input in self.__threat_label_to_color_input.items():
            color = self.GetControlValue(color_input)
            # ida returns sometimes a big int, which is not a real color number. The value must therefore be masked
            if color > WHITE_COLOR and color != idc.DEFCOLOR:
                color = color & WHITE_COLOR  # remove trailing bytes
            self.threat_label_to_color[threat_label] = color
        if fid == idaapi.CB_CANCEL:
            self.canceled = True

        return OKAY_RESPONSE

    @staticmethod
    def create(threat_labels: Set[str]) -> Optional[Dict[str, int]]:
        if not threat_labels:
            return {}
        box = SignatureInspectionColorSelectionBox(threat_labels)
        box.Compile()
        box.Execute()
        if box.canceled:
            return None
        return box.threat_label_to_color
