from typing import List

import idaapi
from PyQt5 import QtWidgets

from threatray_ida_plugin.libs.constants import OKAY_RESPONSE
from threatray_ida_plugin.libs.logger import get_log
from threatray_ida_plugin.models.signature_inspection_result import SignatureInspectionResult
from threatray_ida_plugin.models.signature_inspection_settings import SignatureInspectionSetting
from threatray_ida_plugin.views.utils import get_widget_name

log = get_log()
ADDRESS_COLUMN: int = 0


def _open_function_in_disasm_window(func: idaapi.func_t):
    range_vec = idaapi.rangevec_t()
    range_set = idaapi.rangeset_t()
    ranges = idaapi.get_func_ranges(range_set, func)
    if ranges == idaapi.BADADDR:
        log.warning(f'Failed to get function ranges for {func.start_ea}-{func.end_ea}')
        return
    for i in range(range_set.nranges()):
        range_vec.push_back(range_set.getrange(i))

    # close previous widget to load the function view correctly
    existing_view = idaapi.find_widget('IDA View-Signature Inspection')
    if existing_view:
        idaapi.close_widget(existing_view, idaapi.WCLS_NO_CONTEXT | idaapi.WCLS_DONT_SAVE_SIZE)

    idaapi.open_disasm_window('Signature Inspection', range_vec)


def _handle_item_double_clicked(table_item: QtWidgets.QTableWidgetItem):
    if table_item.column() != ADDRESS_COLUMN:
        return
    try:
        func = idaapi.get_func(int(table_item.text(), 16))
    except (ValueError, TypeError) as e:
        log.error(f'Failed to parse address: {e}')
        return
    _open_function_in_disasm_window(func)


class SignatureInspectionResultView(idaapi.PluginForm):
    def __init__(self, result: List[SignatureInspectionResult], settings: SignatureInspectionSetting):
        super().__init__()
        self.__result: List[SignatureInspectionResult] = result
        self.__settings: SignatureInspectionSetting = settings

    def OnCreate(self, form) -> int:
        main_widget = idaapi.find_widget('IDA View-A')
        idaapi.activate_widget(main_widget, True)
        self.parent = self.FormToPyQtWidget(form)
        self.__create()
        return OKAY_RESPONSE

    def __get_filtered_result(self) -> List[SignatureInspectionResult]:
        if not self.__settings.show_only_functions_with_hit:
            return self.__result
        return [result for result in self.__result if result.threats]

    def __create(self):
        header: List[str] = ['Address', 'Threats', 'Reference Binary File', 'Reference Address']

        layout = QtWidgets.QGridLayout(self.parent)
        table = QtWidgets.QTableWidget(0, len(header))
        table.itemDoubleClicked.connect(_handle_item_double_clicked)
        layout.addWidget(table)

        for row_index, result in enumerate(self.__get_filtered_result()):
            table.insertRow(row_index)
            table.setItem(row_index, ADDRESS_COLUMN, QtWidgets.QTableWidgetItem(hex(result.address)))

            threats = '\n'.join([t.label for t in result.threats])
            reference_binary_files = '\n'.join([t.binary_file for t in result.threats])
            reference_addresses = '\n'.join([hex(t.address) for t in result.threats])
            for column_index, item in enumerate((threats,
                                                 reference_binary_files,
                                                 reference_addresses),
                                                start=1):
                table.setItem(row_index, column_index, QtWidgets.QTableWidgetItem(item))

        table.setHorizontalHeaderLabels(header)
        table.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)
        table.resizeRowsToContents()

    def Show(self) -> idaapi.PluginForm.Show:
        return idaapi.PluginForm.Show(self,
                                      get_widget_name('Signature Inspection Results'),
                                      options=(idaapi.PluginForm.WOPN_PERSIST
                                               | idaapi.PluginForm.WCLS_SAVE
                                               | idaapi.PluginForm.WOPN_MENU
                                               | idaapi.PluginForm.WOPN_RESTORE
                                               | idaapi.PluginForm.WOPN_TAB),
                                      )

    @staticmethod
    def create(result: List[SignatureInspectionResult], settings: SignatureInspectionSetting):
        f = SignatureInspectionResultView(result, settings)
        f.Show()
