from typing import Optional

import idaapi

from threatray_ida_plugin.libs.constants import OKAY_RESPONSE
from threatray_ida_plugin.libs.logger import get_log
from threatray_ida_plugin.models.signature_inspection_settings import SignatureInspectionSetting

log = get_log()

TEXT = r"""BUTTON YES* Select
BUTTON NO NONE
BUTTON CANCEL Cancel
Signature Inspection Settings

Please select your desired setting for the Signature Inspection query:
{form_change}
<Show only functions with signature hits in the result view:{show_only_functions_with_hit}>
<Color functions with signature hits and their xrefs (might overwrite existing colors):{color_functions_with_hit}>
<Manually select the signature hits that are to be colored:{manually_select_signature_hits}>{settings}>
"""


class SignatureInspectionSettingBox(idaapi.Form):
    def __init__(self):
        self.canceled: bool = False
        self.__show_only_functions_with_hit_default: bool = True
        self.__color_functions_with_hit_default: bool = True
        self.__manually_select_signature_hits_to_be_colored_default: bool = False
        self.selected_show_only_functions_with_hit: bool = self.__show_only_functions_with_hit_default
        self.selected_color_functions_with_hit: bool = self.__color_functions_with_hit_default
        self.selected_manually_select_signature_hits_to_be_colored: bool = \
            self.__manually_select_signature_hits_to_be_colored_default

        idaapi.Form.__init__(
            self,
            TEXT,
            {
                'form_change': idaapi.Form.FormChangeCb(self.on_form_change),
                'settings': idaapi.Form.ChkGroupControl((
                    'show_only_functions_with_hit',
                    'color_functions_with_hit',
                    'manually_select_signature_hits'
                )),
            }
        )

    def on_form_change(self, fid):
        self.selected_show_only_functions_with_hit = bool(self.GetControlValue(self.show_only_functions_with_hit))
        self.selected_color_functions_with_hit = bool(self.GetControlValue(self.color_functions_with_hit))
        self.selected_manually_select_signature_hits_to_be_colored = bool(
            self.GetControlValue(self.manually_select_signature_hits))

        if not self.selected_color_functions_with_hit:
            self.EnableField(self.manually_select_signature_hits, False)
        else:
            self.EnableField(self.manually_select_signature_hits, True)

        if fid == idaapi.CB_CANCEL:
            self.canceled = True

        return OKAY_RESPONSE

    def __set_defaults(self):
        self.show_only_functions_with_hit.checked = self.__show_only_functions_with_hit_default
        self.color_functions_with_hit.checked = self.__color_functions_with_hit_default
        self.manually_select_signature_hits.checked = self.__manually_select_signature_hits_to_be_colored_default

    def Compile(self):
        super().Compile()
        self.__set_defaults()

    @staticmethod
    def create() -> Optional[SignatureInspectionSetting]:
        box = SignatureInspectionSettingBox()
        box.Compile()
        box.Execute()
        if box.canceled:
            return None
        return SignatureInspectionSetting(
            show_only_functions_with_hit=box.selected_show_only_functions_with_hit,
            color_functions_with_hit=box.selected_color_functions_with_hit,
            manually_select_signature_hits_to_be_colored=box.selected_manually_select_signature_hits_to_be_colored
        )
