from dataclasses import dataclass
from typing import Optional

import idaapi

from threatray_ida_plugin.libs.constants import OKAY_RESPONSE
from threatray_ida_plugin.libs.logger import get_log

log = get_log()

TEXT = r"""BUTTON YES* Ok
BUTTON NO NONE
BUTTON CANCEL Cancel
Signature thresholds

{form_change}
Signature does not exist and will be created.
Please adjust the following thresholds:
<Threshold malicious:{threshold_malicious}>
<Threshold suspicious:{threshold_suspicious}>
"""


@dataclass(frozen=True)
class SignatureCreationThresholdResponse:
    threshold_suspicious: int
    threshold_malicious: int


class SignatureCreationThresholdBox(idaapi.Form):
    def __init__(self, number_of_selected_functions: int):
        self.canceled: bool = False
        self.selected_threshold_suspicious: int = max(int(number_of_selected_functions * 0.1), 1)
        self.selected_threshold_malicious: int = max(int(number_of_selected_functions * 0.2), 1)

        idaapi.Form.__init__(
            self,
            TEXT,
            {
                'form_change': idaapi.Form.FormChangeCb(self.on_form_change),
                'threshold_suspicious': idaapi.Form.NumericInput(value=self.selected_threshold_suspicious,
                                                                 tp=idaapi.Form.FT_DEC),
                'threshold_malicious': idaapi.Form.NumericInput(value=self.selected_threshold_malicious,
                                                                tp=idaapi.Form.FT_DEC),
            }
        )

    def on_form_change(self, fid):
        try:
            threshold_suspicious = int(self.GetControlValue(self.threshold_suspicious))
            threshold_malicious = int(self.GetControlValue(self.threshold_malicious))
            if threshold_suspicious > threshold_malicious:
                raise ValueError('Threshold suspicious must be less than or equal to threshold malicious.')
            if threshold_suspicious < 1 or threshold_malicious < 1:
                raise ValueError('Thresholds must be a positive integer.')
            self.selected_threshold_suspicious = threshold_suspicious
            self.selected_threshold_malicious = threshold_malicious
        except (ValueError, TypeError) as e:
            log.error(f'Invalid thresholds provided: {e}')

        # we update the thresholds in the UI in case a float number was provided or an error has happened
        self.SetControlValue(self.threshold_suspicious, self.selected_threshold_suspicious)
        self.SetControlValue(self.threshold_malicious, self.selected_threshold_malicious)

        if fid == idaapi.CB_CANCEL:
            self.canceled = True
        return OKAY_RESPONSE

    @staticmethod
    def create(number_of_selected_functions: int) -> Optional[SignatureCreationThresholdResponse]:
        box = SignatureCreationThresholdBox(number_of_selected_functions)
        box.Compile()
        box.Execute()
        if box.canceled:
            return None
        return SignatureCreationThresholdResponse(threshold_suspicious=box.selected_threshold_suspicious,
                                                  threshold_malicious=box.selected_threshold_malicious)
