__author__ = 'threatray'
__email__ = 'support@threatray.com'
__version__ = '1.3.0'
__ida_version__ = 'IDA PRO 8.3'

from dataclasses import dataclass

import idaapi

from threatray_ida_plugin.application.function_retrohunt_service import FunctionRetrohuntService
from threatray_ida_plugin.application.signature_inspection_service import SignatureInspectionService
from threatray_ida_plugin.application.signature_management_service import SignatureManagementService
from threatray_ida_plugin.handlers.function_retrohunt.function_retrohunt_disassembly_view_handler import \
    FunctionRetrohuntDisassemblyViewHook, FunctionRetrohuntDisassemblyViewHandler
from threatray_ida_plugin.handlers.function_retrohunt.function_retrohunt_handler import FunctionRetrohuntHandler, \
    FunctionRetrohuntHook
from threatray_ida_plugin.handlers.signature_inspection_handler import SignatureInspectionHandler
from threatray_ida_plugin.handlers.signature_management_handler import SignatureManagementHandler, SignatureManagerHook
from threatray_ida_plugin.libs.api.api import API
from threatray_ida_plugin.libs.api.api_error import APISetupError
from threatray_ida_plugin.libs.constants import PLUGIN_ID, PLUGIN_NAME, PLUGIN_LOCATION
from threatray_ida_plugin.libs.logger import get_log
from threatray_ida_plugin.views.help_box import HelpBox

log = get_log()


@dataclass(frozen=True)
class PluginDependencies:
    function_retrohunt_service: FunctionRetrohuntService
    signature_inspection_service: SignatureInspectionService
    signature_management_service: SignatureManagementService


def create_dependencies() -> PluginDependencies:
    api = API()
    return PluginDependencies(
        function_retrohunt_service=FunctionRetrohuntService(api),
        signature_inspection_service=SignatureInspectionService(api),
        signature_management_service=SignatureManagementService(api)
    )


class ThreatrayPlugin(idaapi.plugin_t):
    comment = 'Threatray Plugin'
    help = 'Brings threatray features into IDA'
    wanted_name = PLUGIN_NAME  # name in the menu
    wanted_hotkey = ''  # TBD
    flags = idaapi.PLUGIN_KEEP

    def init(self) -> int:
        try:
            dependencies = create_dependencies()
        except APISetupError as e:
            log.error(f'Failed to load {PLUGIN_ID} plugin: {str(e)}')
            return idaapi.PLUGIN_UNL

        idaapi.create_menu(f'{PLUGIN_ID}:main_menu', PLUGIN_NAME, PLUGIN_LOCATION)
        plugin_menu = f'{PLUGIN_LOCATION}{PLUGIN_NAME}/'
        SignatureInspectionHandler(dependencies.signature_inspection_service).register_and_attach_to_menu(plugin_menu)

        FunctionRetrohuntHandler(dependencies.function_retrohunt_service).register_and_attach_to_menu(plugin_menu)
        self._function_retrohunt_hook = FunctionRetrohuntHook()
        self._function_retrohunt_hook.hook()

        # this handler is only used by the UI hook for the disasm context menu
        FunctionRetrohuntDisassemblyViewHandler(dependencies.function_retrohunt_service).register()
        self._function_retrohunt_disassembly_view_hook = FunctionRetrohuntDisassemblyViewHook()
        self._function_retrohunt_disassembly_view_hook.hook()

        log.info(f'{PLUGIN_ID} plugin loaded.')
        return idaapi.PLUGIN_KEEP

    def run(self, arg):
        HelpBox.create(__version__, __email__)

    def term(self):
        pass


def PLUGIN_ENTRY():
    return ThreatrayPlugin()
